var uitl = require('util');
var fs = require('fs');
var xml2js = require('xml2js');
var xlsx = require('xlsx');
var mustache = require('mustache');
var exec = require('child_process').execFile;
var protobuf = require('protocol-buffers');
var colors = require('colors');
var package = require('./package.json');
var program = require('commander');

program
  .version(package.version)
  .description('The tool convert excel to protobuf serializable file and output .proto & .cs files')
  .usage('[options]')
  .option('-i, --input [type]', 'Input directory for excels and config.xml file', checkPathExists)
  .option('-o, --output [type]', 'Output directory', checkPathExists)
  .parse(process.argv);

if(!program.input || !program.output) {
  console.log('All options are required, or using option -h look help.'.yellow);
  return;
}
//==========================
// global variables
//==========================
var protoNamespace = 'ProtoMessages'
var outputDir = program.output + '/generated';
var outputProto = outputDir + '/ProtoMessages.proto'
var outputCs = outputDir + '/ProtoMessages.cs'
var outputSerialDir = outputDir + '/serializable';
var protoGenExe = __dirname + '/proto-gen/protogen.exe';
var inputExcelDir = program.input;
var xmlFilePath = inputExcelDir + '/config.xml';

// check input directory
if(!fs.existsSync(xmlFilePath)) {
  console.log((xmlFilePath + ' is required').bgRed);
  return;
}

// Create output folders
if (fs.existsSync(outputDir)) deleteFolderRecursive(outputDir);
fs.mkdirSync(outputDir);
if (!fs.existsSync(outputSerialDir)) fs.mkdirSync(outputSerialDir);

parseXml(config => {
  let tpl = fs.readFileSync(__dirname + '/template.mustache', 'utf-8');
  fs.writeFileSync(`${outputProto}`, mustache.render(tpl, config));
  exec(protoGenExe, [`-i:${outputProto}`, `-o:${outputCs}`], (err, result) => {
    if (err === null) {
      console.log('.proto, .cs files generated'.green);
      parseXlsx(config, messageSetDatas => { serialize(messageSetDatas) });
    } else {
      console.log(err);
    }
  });
});

function parseXlsx(config, OnXlsxProcessed) {
  let messageSetDatas = [];

  config.workbooks.forEach(wbCfg => {
    console.log(`read ${wbCfg.name}...`.green);
    let workbook = xlsx.readFile(`${inputExcelDir}/${wbCfg.name}`);

    wbCfg.sheets.forEach(shtCfg => {
      console.log(`parse ${shtCfg.name}`.green);
      let worksheet = workbook.Sheets[shtCfg.name];
      let range = worksheet['!range'];

      let messageSetData = { name: `${shtCfg.msgName}`, set: [] };
      for (let r = range.s.r + 1; r < range.e.r; r++) { // exclude header row
        let messageFields = [];
        for (let c = range.s.c; c < range.e.c; c++) {
          let address = String.fromCharCode(65 + c) + (r + 1);
          let field = shtCfg.fields[c];
          if (!field) {
            console.log(`Error: Current col number is ${c}, please check sheet ${shtCfg.name}`.bgRed);
            return;
          }
          messageFields.push({ fieldName: field.fieldName, fieldVal: worksheet[address].v });
        }
        messageSetData.set.push(messageFields);
      }
      messageSetDatas.push(messageSetData);
    });
  });
  OnXlsxProcessed(messageSetDatas);
}

function serialize(messageSetDatas) {
  // serialize messages
  let messages = protobuf(fs.readFileSync(`${outputProto}`));

  messageSetDatas.forEach(messageSetData => {
    msgSetStr = ''
    messageSetData.set.forEach(messageFields => {
      msgStr = '{'
      messageFields.forEach(field => {
        msgStr += `${field.fieldName}:'${field.fieldVal}',`;
      })
      msgStr += '},'
      msgSetStr += msgStr;
    });
    
    let encodeCmd = `messages.${messageSetData.name}Set.encode({${messageSetData.name}: [${msgSetStr}] })`;
    fs.writeFile(`${outputSerialDir}/${messageSetData.name}Set.bytes`, eval(encodeCmd), (err, result) => {
      if (err == null) {
        console.log(`${messageSetData.name}.bytes genrated...`.green);
      } else {
        console.log(err); ``
      }
    });
  });
}

function parseXml(OnXmlProcessComplete) {
  let config = { namepace: protoNamespace, workbooks: [] };
  let parser = new xml2js.Parser();
  let data = fs.readFileSync(xmlFilePath);

  parser.parseString(data, function (err, result) {
    // console.log(result.config.workbook[0].$.name);
    result.config.workbook.forEach(wb => {
      let workbook = { name: wb.$.name, sheets: [] }
      wb.sheet.forEach(st => {
        let sheet = {
          name: st.$.name,
          msgName: st.$.msgName,
          fields: []
        };

        st.cell.forEach((cell, index) => {
          let field = {
            fieldType: cell.$.fieldType,
            fieldName: cell.$.fieldName,
            fieldNumber: (index + 1)
          }
          sheet.fields.push(field);
        });
        workbook.sheets.push(sheet);
      });
      config.workbooks.push(workbook);
    });
  });
  OnXmlProcessComplete(config);
}

function deleteFolderRecursive(path) {
  if (fs.existsSync(path)) {
    fs.readdirSync(path).forEach(function (file, index) {
      var curPath = path + "/" + file;
      if (fs.lstatSync(curPath).isDirectory()) { // recurse
        deleteFolderRecursive(curPath);
      } else { // delete file
        fs.unlinkSync(curPath);
      }
    });
    fs.rmdirSync(path);
  }
};

function checkPathExists(filepath){
  if(!fs.existsSync(filepath)) console.log((filepath + ' path dose not exists').red);
  return filepath;
}
